@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 #
-#   Copyright (c) 2014-2020 eGauge Systems LLC
+#   Copyright (c) 2014-2021 eGauge Systems LLC
 #	1644 Conestoga St, Suite 2
 #	Boulder, CO 80301
 #	voice: 720-545-9767
@@ -76,6 +76,8 @@ import sys
 import tempfile
 import xml.etree.ElementTree as ET
 
+import sexpdata
+
 # Please replace 'epic-sample' with the name of your Django project:
 os.environ.setdefault('DJANGO_SETTINGS_MODULE',
                       'epic-sample.epic-sample.settings')
@@ -92,6 +94,16 @@ from epic.models import Part, format_part_number, Assembly_Item, strchoice, \
 
 MY_TOOL_NAME = 'kicad-to-epic-bom'
 
+KICAD_FOOTPRINT = sexpdata.Symbol('footprint')
+KICAD_PATH = sexpdata.Symbol('path')
+KICAD_PROP = sexpdata.Symbol('property')
+KICAD_PROP_EP = 'EP'				# eGauge part number
+KICAD_PROP_FOOTPRINT = 'Footprint'
+KICAD_PROP_REFDES = 'Reference'
+KICAD_SYMBOL = sexpdata.Symbol('symbol')
+KICAD_SYMBOL_INSTANCES = sexpdata.Symbol('symbol_instances')
+KICAD_UUID = sexpdata.Symbol('uuid')
+
 manufacturer = cfg.EPIC_MANUFACTURER
 
 class Component:
@@ -149,7 +161,13 @@ def save_assembly(assembly_part, bom):
              assembly_part.mfg, assembly_part.mfg_pn, len(bom)),
           file=sys.stderr)
 
-def update_comp(comp):
+def legacy_update_comp(comp):
+    '''Check the footprint of a component in legacy format and update it
+    if necessary.  COMP is the component to update in the form of a
+    list of text lines.  Returns True if the component was updated,
+    False otherwise.
+
+    '''
     footprint_field_index = part_number = None
     has_changed = False
     refdes = ''
@@ -204,35 +222,209 @@ def update_comp(comp):
                     has_changed = False
     return has_changed
 
-def update_schematic(dirname, filename):
+def legacy_update_schematic(path, out):
+    '''Update the component footprints in a schematic in legacy format
+    (pre KiCad v6) as needed.  PATH is the filename path of the
+    schematic, OUT is a temporary output file to write the updated
+    schematic file to.  Returns True if the schematic file changed,
+    False otherwise.
+
+    '''
+    has_changes = False
+    curr_comp = []
+    with io.open(path, 'r', encoding='utf-8') as schem_file:
+        inside_comp = False
+        for line in schem_file:
+            if line.rstrip() == '$Comp':
+                inside_comp = True
+                curr_comp.append(line)
+            elif line.rstrip() == '$EndComp':
+                inside_comp = False
+                curr_comp.append(line)
+                has_changes |= legacy_update_comp(curr_comp)
+                for comp_line in curr_comp:
+                    out.write(comp_line)
+                curr_comp = []
+            elif inside_comp:
+                curr_comp.append(line)
+            else:
+                out.write(line)
+    return has_changes
+
+def modern_update_symbol(symbol, updated_syms):
+    '''Check the footprint of a schematic symbol and update it if
+    necessary.  SYMBOL is the s-expression of the symbol to update.
+    UPDATED_SYMS is a dictionary of all symbol UUIDs whose footprints
+    have been updated.  The value of each entry is the new footprint
+    for that symbol.  Returns True if the symbol was updated, False
+    otherwise.
+
+    '''
+    has_changed = False
+    part_number = footprint_sexp = None
+    refdes = ''
+    for sexp in symbol:
+        if not isinstance(sexp, list):
+            continue
+        if sexp[0] == KICAD_UUID:
+            uuid = sexp[1].value()
+        elif sexp[0] == KICAD_PROP:
+            prop_name = sexp[1]
+            if prop_name == KICAD_PROP_EP:
+                try:
+                    part_number = int(sexp[2])
+                except ValueError:
+                    print('%s: warn: invalid EPIC part number "%s"' %
+                          (MY_TOOL_NAME, sexp[2]), file=sys.stderr)
+            elif prop_name == KICAD_PROP_FOOTPRINT:
+                footprint_sexp = sexp
+            elif prop_name == KICAD_PROP_REFDES:
+                refdes = sexp[2] + ' '
+
+    if part_number is not None:
+        try:
+            part = Part.objects.get(pk=part_number)
+        except ObjectDoesNotExist:
+            print('%s: error: Part %s does not exist' %
+                  (MY_TOOL_NAME, format_part_number(part_number)),
+                  file=sys.stderr)
+            return False
+        footprint = part.footprint
+        if not footprint:
+            footprint = ''
+            if part.mounting not in [Part.MOUNTING_CHASSIS, Part.MOUNTING_FREE]:
+                print('%s: warn: EPIC part %s has no footprint defined' %
+                      (MY_TOOL_NAME, format_part_number(part.id)),
+                      file=sys.stderr)
+
+        if footprint_sexp is None:
+            print('%s: error: Footprint property missing in symbol!'
+                  % (MY_TOOL_NAME), file=sys.stderr)
+            return False
+
+        old_footprint = footprint_sexp[2]
+        if old_footprint != footprint:
+            footprint_sexp[2] = footprint
+            if not old_footprint:
+                msg = 'set'
+            else:
+                msg = 'changed from %s' % old_footprint
+            print('%s: warn: %sfootprint %s to %s.'
+                  % (MY_TOOL_NAME, refdes, msg, footprint))
+            updated_syms[uuid] = footprint
+            has_changed = True
+    return has_changed
+
+def update_symbol_instance(symbol, new_footprint):
+    '''Update a symbol instance whose footprint has changed.  SYMBOL is
+    the s-expression of the symbol to update.  NEW_FOOTPRINT is the
+    name of footprint to set for the symbol.
+
+    '''
+    for sexp in symbol:
+        if not isinstance(sexp, list):
+            continue
+        if sexp[0] == KICAD_FOOTPRINT:
+            sexp[1] = new_footprint
+            return
+    symbol.append([KICAD_FOOTPRINT, new_footprint])
+
+def update_symbol_instances(symbol, updated_syms):
+    '''Check SYMBOL's UUID is in UPDATED_SYMS and, if so, update its
+    footprint with the name given as the value of the entry in
+    UPDATED_SYMS.  Returns True if the symbol was updated, False
+    otherwise.
+
+    '''
+    changed = False
+    for sexp in symbol:
+        if not isinstance(sexp, list):
+            continue
+        if sexp[0] == KICAD_PATH:
+            path = sexp[1].split('/')
+            sym_path = path[-1]
+            if sym_path in updated_syms:
+                update_symbol_instance(sexp, updated_syms[sym_path])
+                changed = True
+    return changed
+
+def modern_update_schematic(path, out, updated_syms):
+    '''Update the schematic symbol footprints in a schematic in modern
+    format (post KiCad v5) as needed. PATH is the filename path of the
+    schematic, OUT is a temporary output file to write the updated
+    schematic file to.  UPDATED_SYMS is a dictionary of all symbol
+    UUIDs whose footprints have been updated.  The value of each entry
+    is the new footprint for that symbol.  Returns True if the
+    schematic file changed, False otherwise.
+
+    '''
+    changed = False
+    with io.open(path, 'r', encoding='utf-8') as schem_file:
+        schematic = sexpdata.load(schem_file)
+        for sexp in schematic:
+            if isinstance(sexp, list) and sexp[0] == KICAD_SYMBOL:
+                if modern_update_symbol(sexp, updated_syms):
+                    changed = True
+    if changed:
+        sexpdata.dump(schematic, out)
+    return changed
+
+def update_schematic(dirname, filename, updated_syms):
     '''Update schematic file FILENAME residing in directory DIRNAME with
-    footprints from the EPIC database.
+    footprints from the EPIC database.  UPDATED_SYMS is a dictionary
+    into which are entered all symbol UUIDs whose footprints have been
+    updated.  The value of each entry is the new footprint for that
+    symbol.  Returns True if the schematic was updated, False
+    otherwise.
 
     '''
     schem_path = os.path.join(dirname, filename)
-    has_changes = False
-    with tempfile.NamedTemporaryFile(dir=dirname) as out:
-        with io.open(schem_path, 'r', encoding='utf-8') as schem_file:
-            inside_comp = False
-            for line in schem_file:
-                if line.rstrip() == '$Comp':
-                    inside_comp = True
-                    curr_comp = [line]
-                elif line.rstrip() == '$EndComp':
-                    inside_comp = False
-                    curr_comp.append(line)
-                    has_changes |= update_comp(curr_comp)
-                    for comp_line in curr_comp:
-                        out.write(comp_line.encode('utf-8'))
-                elif inside_comp:
-                    curr_comp.append(line)
-                else:
-                    out.write(line.encode('utf-8'))
+    with tempfile.NamedTemporaryFile(dir=dirname, mode='w',
+                                     encoding='utf-8') as out:
+        _, ext = os.path.splitext(schem_path)
+        if ext == '.sch':
+            has_changes = legacy_update_schematic(schem_path, out)
+        else:
+            has_changes = modern_update_schematic(schem_path, out,
+                                                  updated_syms)
         if has_changes:
             os.rename(schem_path, schem_path + '~')
             os.link(os.path.join(dirname, out.name), schem_path)
-            print('%s: info: %s updated.  Use kicad to update net list and PCB'
+            print('%s: info: %s updated.  Use kicad to update PCB.'
                   % (MY_TOOL_NAME, schem_path))
+    return has_changes
+
+def update_schematic_sym_list(dirname, filename, has_backup, updated_syms):
+    '''Update the footprints in the symbol-instances list of the top-level
+    schematic of a KiCad v6 or newer schematic file.  DIRNAME is the
+    directory containing the schematic and FILENAME is the filename
+    within that directory.  HAS_BACKUP is True if the schematic file
+    already has been backed up (due to an earlier change).
+    UPDATED_SYMS is a dictionary of all symbol UUIDs whose footprints
+    have been updated.  The value of each entry is the new footprint
+    for that symbol.
+
+    '''
+    schem_path = os.path.join(dirname, filename)
+    with tempfile.NamedTemporaryFile(dir=dirname, mode='w',
+                                     encoding='utf-8') as out, \
+         io.open(schem_path, 'r', encoding='utf-8') as schem_file:
+        schematic = sexpdata.load(schem_file)
+        changed = False
+        for sexp in schematic:
+            if isinstance(sexp, list) \
+               and sexp[0] == KICAD_SYMBOL_INSTANCES:
+                if update_symbol_instances(sexp, updated_syms):
+                    changed = True
+
+        if changed:
+            sexpdata.dump(schematic, out)
+            if not has_backup:
+                os.rename(schem_path, schem_path + '~')
+            os.link(os.path.join(dirname, out.name), schem_path)
+            if not has_backup:
+                print('%s: info: %s updated.  Use kicad to update PCB.'
+                      % (MY_TOOL_NAME, schem_path))
 
 def bom_append(bom, component):
     if component.part_id not in bom:
@@ -383,8 +575,15 @@ def main():
                 sources.append(title_block.find('source').text)
 
     if not args.no_update_footprints:
+        updated_syms = {}
+        has_backup = False
         for schematic in sources:
-            update_schematic(source_path, schematic)
+            if update_schematic(source_path, schematic, updated_syms) \
+               and schematic == sources[0]:
+                has_backup = True
+        if sources and updated_syms:
+            update_schematic_sym_list(source_path, sources[0], has_backup,
+                                      updated_syms)
 
     assembly_name = 'bom:' + schematic_name + revision
     if flavor is not None:
@@ -487,7 +686,7 @@ def main():
                    src_part_dict[best_part.id][1]))
 
         if c.value != c.part.val:
-            print('%s: warn: %s has value %s but part %s has value %s'
+            print('%s: warn: %s has value "%s" but part %s has value "%s"'
                   % (MY_TOOL_NAME, c.refdes, c.value,
                      format_part_number(part_id), c.part.val), file=sys.stderr)
         c.footprint = c.part.footprint
